import numpy as np
from PIL import Image

from cube_map import CubeMap
from equirect import Equirect
from healpix import HealPix

PROJECTION_BY_NAME = {projection.__name__: projection for projection in {
    CubeMap,
    Equirect,
    HealPix,
}}


def convert_sphere(sphere, projection):
    return (
        sphere if isinstance(sphere, projection)
        else projection.from_sphere(sphere)
    )


def load_sphere(file_name, projection):
    image = Image.open(file_name).convert('RGBA')
    array = np.array(image, dtype=np.float32) / 255
    if projection is not None:
        if projection.check_image(array) != 0:
            raise RuntimeError(
                'The contents of %s are not in %s format.'
                % (file_name, projection.__name__)
            )
    else:
        formats = detect_format(array)
        if len(formats) < 1:
            raise LookupError('Unknown format.')
        if len(formats) > 1:
            raise LookupError('Ambiguous format: %s.' % formats)
        projection = formats[0]
    sphere = projection.from_image(array)
    return sphere


def save_sphere(file_name, sphere):
    array = np.array(sphere.to_image() * 255, dtype=np.uint8)
    image = Image.fromarray(array)
    image.save(file_name)


def detect_format(image):
    compatible_projections = []
    for projection in PROJECTION_BY_NAME.itervalues():
        if projection.check_image(image) == 0:
            compatible_projections.append(projection)
    return compatible_projections


def get_format(name):
    return PROJECTION_BY_NAME.get(name)



import copy
import pyglet
from pyglet.gl import *
from pyrr import Vector3

class SphericalMesh(object):

    def __init__(self, resolution):
        seed = Icosahedron()
        for i in xrange(0, resolution):
            seed = seed.subdivide()
        self.vertices = seed.vertices
        self.indices = []
        for abc in seed.triangles:
            self.indices.extend(abc)
        self.vertex_list = pyglet.graphics.vertex_list_indexed(
            len(self.vertices),
            self.indices,
            ('v3d/static', [
                coord
                for vector in self.vertices
                for coord in vector.normalised
            ]),
        )

    def draw_triangles(self):
        self.vertex_list.draw(gl.GL_TRIANGLES)


class TriangleMesh(object):

    def __init__(self, vertices, edges, triangles):
        self.vertices = vertices
        self.edges = edges
        self.triangles = triangles

    def subdivide(self):
        vertices = copy.deepcopy(self.vertices)
        edge_midpoint = {}
        edges = set()
        for (a, b) in self.edges:
            m_ab = len(vertices)
            vertices.append((vertices[a] + vertices[b]) / 2)
            edge_midpoint[(a, b)] = m_ab
            edge_midpoint[(b, a)] = m_ab
            edges.add((a, m_ab))
            edges.add((b, m_ab))
        triangles = []
        for (a, b, c) in self.triangles:
            m_ab = edge_midpoint[(a, b)]
            m_ac = edge_midpoint[(a, c)]
            m_bc = edge_midpoint[(b, c)]
            triangles.extend([
                (a, m_ab, m_ac),
                (b, m_bc, m_ab),
                (c, m_ac, m_bc),
                (m_ab, m_bc, m_ac),
            ])
            edges.add((m_ab, m_bc))
            edges.add((m_bc, m_ac))
            edges.add((m_ac, m_ab))
        return TriangleMesh(vertices, edges, triangles)


class Icosahedron(TriangleMesh):

    def __init__(self):
        phi = 1.61803398
        fs = [
            lambda a, b: [0, a, b],
            lambda a, b: [a, b, 0],
            lambda a, b: [b, 0, a],
        ]
        vertices = [
            Vector3(f(a, b)).normalised
            for f in fs
            for a in (+phi, -phi)
            for b in (+1, -1)
        ]
        edges = {
            (i, j)
            for j in xrange(0, 12)
            for i in xrange(0, j)
            if (vertices[i] - vertices[j]).length < 1.5
        }
        triangles = [
            (i, j, k)
            if vertices[i] ^ vertices[j] | vertices[k] > 0 else
            (j, i, k)
            for k in xrange(0, 12)
            for j in xrange(0, k)
            for i in xrange(0, j)
            if {(i, j), (i, k), (j, k)}.issubset(edges)
        ]
        self.vertices = vertices
        self.edges = edges
        self.triangles = triangles



import argparse
import numpy as np
from PIL import Image
import pyglet
from pyglet.gl import *
from pyrr import Quaternion, Vector3, Matrix44

import projection as proj
from geom import SphericalMesh
from glsl import Shader

class Sphaira(pyglet.window.Window):

    def __init__(self, **kwargs):
        super(Sphaira, self).__init__(**kwargs)
        self.maximize()
        self.t = 0.0
        self.orientation = Quaternion()
        self.zoom = 2.5
        self.mesh = SphericalMesh(4)
        self.cube_map = None

    def load_file(self, file_name, in_format):
        self.sphere = proj.load_sphere(file_name, projection=in_format)
        in_format = self.sphere.__class__
        print('Loaded input %s from %s.' % (in_format.__name__, file_name))
        self.texture_id = (GLuint * 1)()
        glGenTextures(1, self.texture_id)
        self.sphere.to_gl(self.texture_id[0])
        self.shader = Shader(
            vert=VERTEX_SHADER,
            frag=[
                FRAGMENT_SHADER,
                self.sphere.get_glsl_sampler(),
            ]
        )

    def update(self, dt):
        self.t += dt

    def on_mouse_press(self, x, y, button, modifiers):
        # self.set_exclusive_mouse()
        return

    def on_mouse_release(self, x, y, button, modifiers):
        # self.set_exclusive_mouse(exclusive=False)
        return

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        # rotate on left-drag
        if buttons & 1:
            # the rotation vector is the displacement vector rotated by 90 degrees
            if dx == 0 and dy == 0:
                return
            v = Vector3([dy, -dx, 0])
            # update the current orientation
            self.orientation *= Quaternion.from_axis_rotation(
                -v.normalised,
                v.length * 0.002
            )
        # zoom on right-drag
        if buttons & 4:
            self.zoom += self.zoom * dy*0.01

    def on_draw(self):
        glViewport(0, 0, self.width, self.height)
        self.clear()
        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LESS)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(50, self.width / float(self.height), .01, 100)
        glMatrixMode(GL_MODELVIEW)
        glPushMatrix()
        glTranslatef(0, 0, -0.9)
        m = self.orientation.matrix44
        array = (GLdouble * 16)()
        for i in xrange(4):
            for j in xrange(4):
                array[4*i + j] = m[i,j]
        glMultMatrixd(array)
        glPointSize(1.8)
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        # draw stuff
        self.shader.bind()
        glActiveTexture(GL_TEXTURE0 + self.texture_id[0])
        self.sphere.bind_glsl_texture(self.texture_id[0], self.shader)
        self.mesh.draw_triangles()
        glPopMatrix()


VERTEX_SHADER = '''
#version 130
attribute vec4 vert;
varying vec3 texCoord;
void main()
{
    gl_Position = gl_ModelViewProjectionMatrix * vert;
    texCoord = vert.xyz;
}
'''

FRAGMENT_SHADER = '''
#version 130
varying vec3 texCoord;
vec4 sample(vec3 v);
void main()
{
    gl_FragColor = sample(texCoord);
}
'''


def main():
    parser = argparse.ArgumentParser(
        prog='view',
        description='Sphaira viewer for spherical data.',
    )
    parser.add_argument('-i', '--in_format', help='IN_FORMAT')
    parser.add_argument('input', help='INPUT')
    args = parser.parse_args()
    in_format = proj.get_format(args.in_format)
    config = pyglet.gl.Config(sample_buffers=1, samples=4, double_buffer=True, depth_size=24)
    window = Sphaira(caption='Sphaira Viewer', resizable=True, vsync=True, config=config)
    window.load_file(args.input, in_format)
    pyglet.clock.schedule_interval(window.update, (1.0/60))
    pyglet.app.run()

if __name__ == '__main__':
    main()



# Based on code by Tristam Macdonald.

from pyglet.gl import *
from ctypes import *

class Shader:
    # vert, frag and geom take arrays of source strings
    # the arrays will be concattenated into one string by OpenGL
    def __init__(self, vert = [], frag = [], geom = []):
        # create the program handle
        self.handle = glCreateProgram()
        # we are not linked yet
        self.linked = False

        # create the vertex shader
        self.createShader(vert, GL_VERTEX_SHADER)
        # create the fragment shader
        self.createShader(frag, GL_FRAGMENT_SHADER)
        # the geometry shader will be the same, once pyglet supports the extension
        # self.createShader(frag, GL_GEOMETRY_SHADER_EXT)

        # attempt to link the program
        self.link()

    def createShader(self, strings, type):
        count = len(strings)
        # if we have no source code, ignore this shader
        if count < 1:
            return

        # create the shader handle
        shader = glCreateShader(type)

        # convert the source strings into a ctypes pointer-to-char array, and upload them
        # this is deep, dark, dangerous black magick - don't try stuff like this at home!
        src = (c_char_p * count)(*strings)
        glShaderSource(shader, count, cast(pointer(src), POINTER(POINTER(c_char))), None)

        # compile the shader
        glCompileShader(shader)

        temp = c_int(0)
        # retrieve the compile status
        glGetShaderiv(shader, GL_COMPILE_STATUS, byref(temp))

        # if compilation failed, print the log
        if not temp:
            # retrieve the log length
            glGetShaderiv(shader, GL_INFO_LOG_LENGTH, byref(temp))
            # create a buffer for the log
            buffer = create_string_buffer(temp.value)
            # retrieve the log text
            glGetShaderInfoLog(shader, temp, None, buffer)
            # print the log to the console
            print buffer.value
        else:
            # all is well, so attach the shader to the program
            glAttachShader(self.handle, shader);

    def link(self):
        # link the program
        glLinkProgram(self.handle)

        temp = c_int(0)
        # retrieve the link status
        glGetProgramiv(self.handle, GL_LINK_STATUS, byref(temp))

        # if linking failed, print the log
        if not temp:
            #   retrieve the log length
            glGetProgramiv(self.handle, GL_INFO_LOG_LENGTH, byref(temp))
            # create a buffer for the log
            buffer = create_string_buffer(temp.value)
            # retrieve the log text
            glGetProgramInfoLog(self.handle, temp, None, buffer)
            # print the log to the console
            print buffer.value
        else:
            # all is well, so we are linked
            self.linked = True

    def bind(self):
        # bind the program
        glUseProgram(self.handle)

    def unbind(self):
        # unbind whatever program is currently bound - not necessarily this program,
        # so this should probably be a class method instead
        glUseProgram(0)

    # upload a floating point uniform
    # this program must be currently bound
    def uniformf(self, name, *vals):
        # check there are 1-4 values
        if len(vals) in range(1, 5):
            # select the correct function
            { 1 : glUniform1f,
                2 : glUniform2f,
                3 : glUniform3f,
                4 : glUniform4f
                # retrieve the uniform location, and set
            }[len(vals)](glGetUniformLocation(self.handle, name), *vals)

    # upload an integer uniform
    # this program must be currently bound
    def uniformi(self, name, *vals):
        # check there are 1-4 values
        if len(vals) in range(1, 5):
            # select the correct function
            { 1 : glUniform1i,
                2 : glUniform2i,
                3 : glUniform3i,
                4 : glUniform4i
                # retrieve the uniform location, and set
            }[len(vals)](glGetUniformLocation(self.handle, name), *vals)

    # upload a uniform matrix
    # works with matrices stored as lists,
    # as well as euclid matrices
    def uniform_matrixf(self, name, mat):
        # obtian the uniform location
        loc = glGetUniformLocation(self.Handle, name)
        # uplaod the 4x4 floating point matrix
        glUniformMatrix4fv(loc, 1, False, (c_float * 16)(*mat))
